/*
 * Copyright (c) 1998 Qualcomm Incorporated. All rights reserved.
 * 
 * Description : 
 *   maillock() and mailunlock() try to emulate the SysV functions
 *   to create the user.lock file.
 *   call touchlock() to update the lock.
 * Caveats :
 *   mailunlock() will only remove the lockfile created from previous
 *   maillock(). maillock() called for different users with out the
 *   intermediate calls for mailunlock() would leave the initial locks.
 *   Process termination would not clean the .lock files.
 * 
 * Revisions :
 *   09/02/98 [py]
 *        Created.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#if HAVE_SYS_FILE_H
# include <sys/file.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <flock.h>
#include <netdb.h>
#include <maillock.h>
#include <popper.h>


#ifndef MAILLOCK

/* *** MACROS *** */
/* 5 minutes */
#define LOCK_LIFE      300
#define SLEEP_INTERVAL 5
#ifndef MAXPATHLEN
#define MAXPATHLEN     FILENAME_MAX
#endif
#define BUF_SIZE       (MAXHOSTNAMELEN + 32)

/* *** PRIVATE GLOBALS *** */
static char pszLock[MAXPATHLEN] = { 0 };   /* Current lock path is saved */

/* *** PRIVATE FUNCTIONS *** */


/*
 * Function   : maillock
 * Descrption : Creates .lock file in the spool directory for the mailbox of the
 *              user.
 *              Every maillock should be paired with mailunlock before a
 *              subsequent call to maillock() is made.
 * Parameters : user - user id.
 *              retrycnt - Retry Count. Between retries there is a sleep of 5 second
 *              times the retrycount.
 * Returns    : L_SUCCESS on success. 
 *
 * Caveats    : Clock drift on Shared FS not accounted.
 */

int maillock( user, retrycnt )
char *user;
int  retrycnt;
{
    int         fd = -1;
    struct stat byNameStat;
    int         attemptcnt = 0;
    time_t      curTime;
    int         nRet = L_SUCCESS;
#if defined(NO_ATOMIC_OPEN)
    char        *pTmp;
#endif
    /*
     * Get the path to mail box.
     * Append the .lock to name.
     */
    /* <todo> Use genpath() to obtain the path to mbox. </todo> */
    if( genpath( user, pszLock ) == -1 ) {
	/* LOG */
	return L_ERROR;
    }
    strcat( pszLock, ".lock" );

    /*
     * Verify if lockfile is a symbolic link, or a hard link.
     * Eventhough the link wouldn't be followed with O_CREAT | O_EXCL
     * we better do it.
     */
    if ( lstat(pszLock,&byNameStat) == -1 ) {
	if( errno != ENOENT ) {
	    /* Log the failure */
	    return L_ERROR;
	}
    }
    else if( (byNameStat.st_mode & S_IFMT) == S_IFLNK  || byNameStat.st_nlink > 1 ){
	/* Log the failure */
	return L_ERROR;
    }
    while ( attemptcnt != retrycnt ) {
	if(attemptcnt) 
	    sleep( attemptcnt * SLEEP_INTERVAL );       /* Sleep between retries */

#if defined(NO_ATOMIC_OPEN)
	while( fd == -1 ) {
	    /* Create a temporary file and link it to lock file */
	    pTmp = tempnam( POP_MAILDIR, "POP" );
	    if( (fd = open( pTmp, O_CREAT | O_EXCL, 0600 )) == -1 ){
		if( errno == EEXIST ) continue;
		else {
		    /* Log the failure */
		    nRet = L_ERROR;
		    attemptcnt = retrycnt;
		    pTmp = 0;
		    break;
		}
	    }
	}
	if ( pTmp && link( pTmp, pszLock ) == -1 ) {
	    if( errno == EEXIST ) {
		curTime = time(0);
		if( stat(pszLock, &byNameStat) != -1 && 
		    curTime - (time_t)(byNameStat.st_mtime) > LOCK_LIFE ) {
		    unlink( pszLock );                  /* Remove the old lock */
		    if( link( pTmp, pszLock ) == -1 ) { /* Place a new lock */
			nRet = L_TMPLOCK;
			/* Log the failure */
		    }
		    attemptcnt = retrycnt;              /* Break out */
		}
		else
		    attemptcnt++;                       /* Try again */
	    }
	    else {
		/* Log the failure */
		nRet =  L_ERROR;
		attemptcnt = retrycnt;
	    }
	}
	else {
	    attemptcnt = retrycnt;
	    unlink( pTmp );
	}
	if( nRet != L_SUCCESS && fd != -1 ){
	    unlink( pTmp );
	    fd = -1;
	}
#else
	/* The following is supposed to be atomic - even on NFS.*/
	if( (fd = open( pszLock, O_WRONLY|O_CREAT|O_EXCL, 0600 )) == -1 ) {
	    switch(errno) {
	    case EACCES:
		/* Keep trying for the lock life*/
		curTime = time(0);
		if( stat(pszLock, &byNameStat) != -1 && 
		    curTime - (time_t)(byNameStat.st_mtime) < LOCK_LIFE ) 
		    attemptcnt++;
		else {
		    attemptcnt = retrycnt;
		    nRet = L_ERROR;
		}
		break;
	    case EEXIST:
		/* Check the lock life */
		curTime = time(0);
		if( stat(pszLock, &byNameStat) != -1 && 
		    curTime - (time_t)(byNameStat.st_mtime) > LOCK_LIFE ) {
		    /* Create the lock file */
		    /* This create could be happening to a link. 
		     * Check the race */
		    if( (fd = open( pszLock, O_WRONLY|O_CREAT|O_TRUNC, 0600 )) 
			== -1 ) {
			/* Log the failure */
			nRet = L_TMPLOCK;
		    }
		    attemptcnt = retrycnt;
		}
		else {
		    attemptcnt++;
		    if( attemptcnt == retrycnt ) nRet = L_MAXTRYS;
		}
		break;
	    default:
		/* Log the failure */
		attemptcnt = retrycnt;
		nRet = L_TMPLOCK;
		break;
	    }
	}
	else {
	    attemptcnt = retrycnt;
	    break;
	}
#endif
    }
    if( fd != -1 ) { /* .lock file created successfully */
	/* Do the lstat again to check if this is a symbolic link to
	 * account for the Race conditions */
	if ( ( lstat(pszLock,&byNameStat) == -1 ) ||
	     ( (byNameStat.st_mode & S_IFMT) == S_IFLNK ) ||
	     ( byNameStat.st_nlink > 1 ) ){
	    /* Log the failure */
	    nRet = L_TMPLOCK;
	}
	/* Lock the file, write some data pid@host and unlock it.*/
	/* Well this is advisory locking. */
	if( flock( fd, LOCK_EX | LOCK_NB ) == -1 ) {
	    nRet = L_MANLOCK; /* Some one seized the lock since 
			       * this process created */
	}
	else {
	    char buf[BUF_SIZE];
	    int  nSize;
	    sprintf(buf, "%d@", getpid() );
	    nSize = strlen(buf);
	    gethostname( buf + nSize, BUF_SIZE - nSize - 1 );
	    buf[BUF_SIZE] = '\0';
	    nSize = strlen(buf);
	    if( write( fd, buf, nSize ) != nSize ) {
		nRet = L_TMPWRITE;
	    }
	    flock( fd, LOCK_UN );
	}
	close( fd );
    }
    return nRet;
}


/*
 * Function   : mailunlock
 * Descrption : Removes the .lock file previously obtained through
 *              maillock().
 * Parameters : NONE.
 * Returns    : L_SUCCESS, L_ERROR - When there wasn't a lock.
 */
#if defined(__STDC__)
void mailunlock(void)
#else
int mailunlock()
#endif
{
    int nRet = L_ERROR;
    if( *pszLock ) {
	unlink( pszLock );
	*pszLock = '\0';
	nRet = L_SUCCESS;
    }
#if defined(__STDC__)
    return;
#else
    return nRet;
#endif
}


/*
 * Function   : touchlock
 * Descrption : Changes access and modification times of the .lock file.
 * Parameters : NONE.
 * Returns    : L_SUCCESS on a success, L_ERROR when there wasn't lock file.
 */
int touchlock()
{
    int nRet = L_ERROR;
    if( *pszLock ) {
	utime( pszLock, NULL ); /* changes access and modification times */
	nRet = L_SUCCESS;
    }
    return L_ERROR;
}












#endif /* MAILLOCK */
